import { Component, Element, Method, Prop, Watch, h, Host, Event } from '@stencil/core';
import { DeckdeckgoChartUtils } from '../deckdeckgo-chart';
import { scaleBand, scaleLinear } from 'd3-scale';
import { max } from 'd3-array';
import { axisBottom, axisLeft } from 'd3-axis';
import 'd3-transition';
export class DeckdeckgoBarChart {
  constructor() {
    /**
     * The line separator use in your csv file
     */
    this.separator = ';';
    /**
     * Set to `true` in case you would like to load (fetch) the data by yourself. Useful in case your data are protected with a token.
     */
    this.customLoader = false;
    /**
     * The margin top of the chart in pixel
     */
    this.marginTop = 32;
    /**
     * The margin bottom of the chart in pixel
     */
    this.marginBottom = 64;
    /**
     * The margin left of the chart in pixel
     */
    this.marginLeft = 32;
    /**
     * The margin right of the chart in pixel
     */
    this.marginRight = 32;
    /**
     * Display multiple graphs and animate the transition between these
     */
    this.animation = false;
    /**
     * Duration of the transition between graphs
     */
    this.animationDuration = 1000;
    this.barDataIndex = 0;
    /**
     * If `false`, no axis y will be draw.
     */
    this.yAxis = true;
    /**
     * Set a minimal value for the y Axis. Useful in case the series of data could contains only zeros.
     */
    this.yAxisMin = 0;
  }
  async componentDidLoad() {
    await this.draw();
  }
  async onSrcChange() {
    await this.draw();
  }
  async onDataChange() {
    await this.draw();
  }
  /**
   * This is the method we are using to refresh the current bar chart when an audience is participating to live vote. It will not redraw the axis but it will draw and animate the bars.
   * @param values
   */
  async updateCurrentBar(values) {
    if (!this.x0 || !this.x1 || !this.y) {
      return;
    }
    if (!this.chartData || this.chartData.length <= 0) {
      return;
    }
    if (this.barDataIndex < 0 || this.barDataIndex >= this.chartData.length) {
      return;
    }
    this.chartData[this.barDataIndex].values = values;
    await this.drawBars(this.barDataIndex, this.animationDuration);
  }
  /**
   * In case you would like to redraw your chart, for example on resize of the window.
   * @param width
   * @param height
   */
  draw(width, height) {
    return new Promise(async (resolve) => {
      if (width > 0) {
        this.width = width;
      }
      if (height > 0) {
        this.height = height;
      }
      if (!this.width || !this.height) {
        resolve();
        return;
      }
      this.svg = DeckdeckgoChartUtils.initSvg(this.el, this.width + this.marginLeft + this.marginRight, this.height + this.marginTop + this.marginBottom);
      this.svg = this.svg.append('g').attr('transform', 'translate(' + (this.marginLeft + this.marginRight) + ',' + (this.marginTop + this.marginBottom) + ')');
      this.barDataIndex = 0;
      this.chartData = this.data;
      if (this.src) {
        this.chartData = await this.fetchData();
      }
      if (!this.chartData || this.chartData === undefined || this.chartData.length <= 0) {
        resolve();
        return;
      }
      await this.firstDraw();
      resolve();
    });
  }
  async firstDraw() {
    await this.initAxis();
    await this.drawAxis();
    await this.initRandomColors();
    await this.drawBars(0, 0);
  }
  async initRandomColors() {
    var _a;
    if (((_a = this.randomColors) === null || _a === void 0 ? void 0 : _a.length) === this.chartData[0].values.length) {
      return;
    }
    this.randomColors = Array.from({ length: this.chartData[0].values.length }, (_v, _i) => Math.floor(Math.random() * 16777215).toString(16));
    this.chartRandomColor.emit(this.randomColors);
  }
  async drawBars(index, animationDuration) {
    await this.initAxisXDomain();
    await this.initAxisYDomain();
    if (this.animation) {
      await this.drawAnimatedBars(index, animationDuration);
    }
    else {
      await this.drawInstantBars();
    }
  }
  initAxisXDomain() {
    return new Promise((resolve) => {
      const xDomains = this.chartData[0].values.map((d) => {
        return d.label;
      });
      if (this.animation) {
        this.initAnimatedAxisX(xDomains);
      }
      else {
        this.initInstantAxisX(xDomains);
      }
      resolve();
    });
  }
  initAxisYDomain() {
    return new Promise((resolve) => {
      const maxValue = max(this.chartData, (category) => {
        return max(category.values, (d) => {
          return d.value;
        });
      });
      this.y.domain([0, Math.max(this.yAxisMin, maxValue)]);
      resolve();
    });
  }
  /**
   * If you are using animation, this method is used to display the next data respectively the next chart.
   */
  async next() {
    await this.prevNext(true);
  }
  /**
   * If you are using animation, this method is used to display the previous data respectively the previous chart.
   */
  async prev() {
    await this.prevNext(false);
  }
  async prevNext(next) {
    if (!this.animation) {
      return;
    }
    if (!this.chartData || this.chartData.length <= 0) {
      return;
    }
    if (next && this.barDataIndex + 1 < this.chartData.length) {
      this.barDataIndex++;
      await this.drawBars(this.barDataIndex, this.animationDuration);
    }
    else if (!next && this.barDataIndex > 0) {
      this.barDataIndex--;
      await this.drawBars(this.barDataIndex, this.animationDuration);
    }
  }
  /**
   * Is animation at the begin of the serie.
   */
  isBeginning() {
    return new Promise((resolve) => {
      if (!this.animation) {
        resolve(true);
        return;
      }
      resolve(this.barDataIndex === 0);
    });
  }
  /**
   * Is animation at the end of the serie.
   */
  isEnd() {
    return new Promise((resolve) => {
      if (!this.animation) {
        resolve(true);
        return;
      }
      resolve(this.barDataIndex === this.chartData.length - 1);
    });
  }
  drawAxis() {
    return new Promise((resolve) => {
      const bottomAxis = axisBottom(this.x0);
      this.svg
        .append('g')
        .attr('class', 'axis axis-x')
        .attr('transform', 'translate(0,' + (this.height - this.marginTop - this.marginBottom) + ')')
        .call(bottomAxis)
        .selectAll('text')
        .attr('transform', 'translate(-8,8)rotate(-45)')
        .style('text-anchor', 'end');
      if (!this.yAxis) {
        resolve();
        return;
      }
      const leftAxis = axisLeft(this.y);
      this.svg.append('g').attr('class', 'axis axis-y').call(leftAxis);
      resolve();
    });
  }
  initAxis() {
    return new Promise(async (resolve) => {
      this.x0 = scaleBand().rangeRound([0, this.width - this.marginLeft - this.marginRight]);
      this.x1 = scaleBand().padding(0.05);
      this.y = scaleLinear().rangeRound([this.height - this.marginTop - this.marginBottom, 0]);
      await this.initAxisXDomain();
      await this.initAxisYDomain();
      resolve();
    });
  }
  initAnimatedAxisX(xDomains) {
    this.x0.domain(xDomains).padding(0.05);
  }
  initInstantAxisX(xDomains) {
    const categoriesNames = this.chartData.map((d) => {
      return d.label;
    });
    this.x0.paddingInner(0.1).domain(categoriesNames);
    this.x1.domain(xDomains).rangeRound([0, this.x0.bandwidth()]);
  }
  drawAnimatedBars(index, animationDuration) {
    return new Promise((resolve) => {
      const section = this.svg.selectAll('rect').data(this.chartData[index].values);
      section
        .enter()
        .append('rect')
        .merge(section)
        .attr('style', (d, i) => this.style(d, i))
        .transition()
        .duration(animationDuration)
        .attr('x', (d) => {
        return this.x0(d.label);
      })
        .attr('y', (d) => {
        return this.y(d.value);
      })
        .attr('width', this.x0.bandwidth())
        .attr('height', (d) => {
        const height = this.height - this.marginTop - this.marginBottom - this.y(d.value);
        return height >= 0 ? height : 0;
      });
      resolve();
    });
  }
  drawInstantBars() {
    return new Promise((resolve) => {
      this.svg
        .append('g')
        .selectAll('g')
        .data(this.chartData)
        .enter()
        .append('g')
        .attr('transform', (d) => {
        return 'translate(' + this.x0(d.label) + ',0)';
      })
        .selectAll('rect')
        .data((d) => {
        return d.values;
      })
        .enter()
        .append('rect')
        .attr('x', (d) => {
        return this.x1(d.label);
      })
        .attr('y', (d) => {
        return this.y(d.value);
      })
        .attr('width', this.x1.bandwidth())
        .attr('height', (d) => {
        const height = this.height - this.marginTop - this.marginBottom - this.y(d.value);
        return height >= 0 ? height : 0;
      })
        .attr('style', (d, i) => this.style(d, i));
      resolve();
    });
  }
  style(d, i) {
    const fillColor = this.randomColors && this.randomColors.length > i ? `, #${this.randomColors[i]}` : '';
    const background = `var(--deckgo-chart-fill-color-${d.key}${fillColor})`;
    return ('fill: ' +
      background +
      '; fill-opacity: var(--deckgo-chart-fill-opacity-' +
      d.key +
      ', 0.8); stroke: var(--deckgo-chart-stroke-' +
      d.key +
      ', ' +
      background +
      '); stroke-width: var(--deckgo-chart-stroke-width-' +
      d.key +
      ', 3px)');
  }
  fetchData() {
    return new Promise(async (resolve) => {
      if (!this.src) {
        resolve([]);
        return;
      }
      if (this.customLoader) {
        this.chartCustomLoad.emit(this.src);
        resolve([]);
        return;
      }
      const response = await fetch(this.src);
      const content = await response.text();
      let results = await this.loadContent(content);
      resolve(results);
    });
  }
  /**
   * If you "manually" load the data, call this method once the text content fetched.
   * @param content
   */
  async postCustomLoad(content) {
    this.chartData = await this.loadContent(content);
    if (!this.chartData || this.chartData === undefined || this.chartData.length <= 0) {
      return;
    }
    await this.firstDraw();
  }
  loadContent(content) {
    return new Promise(async (resolve) => {
      if (!content || content === undefined) {
        resolve([]);
        return;
      }
      const lines = content.split('\n');
      if (!lines || lines.length <= 0) {
        resolve([]);
        return;
      }
      let results = [];
      let keys;
      lines.forEach((line, lineCount) => {
        const values = line.split(this.separator);
        if (values && values.length >= 2) {
          // Title
          if (!keys) {
            keys = Array.apply(null, { length: values.length }).map(Number.call, Number).slice(1);
          }
          let dataValues = [];
          for (let i = 1; i < values.length; i++) {
            const tmp = parseInt(values[i]);
            if (!isNaN(tmp)) {
              dataValues.push({
                key: `${i}`,
                label: keys.length >= i ? `${keys[i - 1]}` : `${i}`,
                value: tmp,
              });
            }
            else if (lineCount === 0 && keys.length >= i) {
              keys[i - 1] = values[i];
            }
          }
          if (dataValues && dataValues.length > 0) {
            results.push({
              label: values[0],
              values: dataValues,
            });
          }
        }
      });
      resolve(results);
    });
  }
  render() {
    return (h(Host, { style: { width: `${this.width}px`, height: `${this.height}px` } },
      h("svg", null)));
  }
  static get is() { return "deckgo-bar-chart"; }
  static get encapsulation() { return "shadow"; }
  static get originalStyleUrls() { return {
    "$": ["deckdeckgo-bar-chart.scss"]
  }; }
  static get styleUrls() { return {
    "$": ["deckdeckgo-bar-chart.css"]
  }; }
  static get properties() { return {
    "width": {
      "type": "number",
      "mutable": true,
      "complexType": {
        "original": "number",
        "resolved": "number",
        "references": {}
      },
      "required": false,
      "optional": false,
      "docs": {
        "tags": [],
        "text": "The width of the chart"
      },
      "attribute": "width",
      "reflect": false
    },
    "height": {
      "type": "number",
      "mutable": true,
      "complexType": {
        "original": "number",
        "resolved": "number",
        "references": {}
      },
      "required": false,
      "optional": false,
      "docs": {
        "tags": [],
        "text": "The height of the chart"
      },
      "attribute": "height",
      "reflect": false
    },
    "src": {
      "type": "string",
      "mutable": false,
      "complexType": {
        "original": "string",
        "resolved": "string",
        "references": {}
      },
      "required": false,
      "optional": false,
      "docs": {
        "tags": [],
        "text": "The path to the source file of the data"
      },
      "attribute": "src",
      "reflect": false
    },
    "separator": {
      "type": "string",
      "mutable": false,
      "complexType": {
        "original": "string",
        "resolved": "string",
        "references": {}
      },
      "required": false,
      "optional": false,
      "docs": {
        "tags": [],
        "text": "The line separator use in your csv file"
      },
      "attribute": "separator",
      "reflect": false,
      "defaultValue": "';'"
    },
    "customLoader": {
      "type": "boolean",
      "mutable": false,
      "complexType": {
        "original": "boolean",
        "resolved": "boolean",
        "references": {}
      },
      "required": false,
      "optional": false,
      "docs": {
        "tags": [],
        "text": "Set to `true` in case you would like to load (fetch) the data by yourself. Useful in case your data are protected with a token."
      },
      "attribute": "custom-loader",
      "reflect": false,
      "defaultValue": "false"
    },
    "marginTop": {
      "type": "number",
      "mutable": false,
      "complexType": {
        "original": "number",
        "resolved": "number",
        "references": {}
      },
      "required": false,
      "optional": false,
      "docs": {
        "tags": [],
        "text": "The margin top of the chart in pixel"
      },
      "attribute": "margin-top",
      "reflect": false,
      "defaultValue": "32"
    },
    "marginBottom": {
      "type": "number",
      "mutable": false,
      "complexType": {
        "original": "number",
        "resolved": "number",
        "references": {}
      },
      "required": false,
      "optional": false,
      "docs": {
        "tags": [],
        "text": "The margin bottom of the chart in pixel"
      },
      "attribute": "margin-bottom",
      "reflect": false,
      "defaultValue": "64"
    },
    "marginLeft": {
      "type": "number",
      "mutable": false,
      "complexType": {
        "original": "number",
        "resolved": "number",
        "references": {}
      },
      "required": false,
      "optional": false,
      "docs": {
        "tags": [],
        "text": "The margin left of the chart in pixel"
      },
      "attribute": "margin-left",
      "reflect": false,
      "defaultValue": "32"
    },
    "marginRight": {
      "type": "number",
      "mutable": false,
      "complexType": {
        "original": "number",
        "resolved": "number",
        "references": {}
      },
      "required": false,
      "optional": false,
      "docs": {
        "tags": [],
        "text": "The margin right of the chart in pixel"
      },
      "attribute": "margin-right",
      "reflect": false,
      "defaultValue": "32"
    },
    "animation": {
      "type": "boolean",
      "mutable": false,
      "complexType": {
        "original": "boolean",
        "resolved": "boolean",
        "references": {}
      },
      "required": false,
      "optional": false,
      "docs": {
        "tags": [],
        "text": "Display multiple graphs and animate the transition between these"
      },
      "attribute": "animation",
      "reflect": false,
      "defaultValue": "false"
    },
    "animationDuration": {
      "type": "number",
      "mutable": false,
      "complexType": {
        "original": "number",
        "resolved": "number",
        "references": {}
      },
      "required": false,
      "optional": false,
      "docs": {
        "tags": [],
        "text": "Duration of the transition between graphs"
      },
      "attribute": "animation-duration",
      "reflect": false,
      "defaultValue": "1000"
    },
    "data": {
      "type": "unknown",
      "mutable": false,
      "complexType": {
        "original": "DeckdeckgoBarChartData[]",
        "resolved": "DeckdeckgoBarChartData[]",
        "references": {
          "DeckdeckgoBarChartData": {
            "location": "import",
            "path": "@deckdeckgo/types"
          }
        }
      },
      "required": false,
      "optional": false,
      "docs": {
        "tags": [],
        "text": "Instead of a source file, source data can also be provided as an array of `DeckdeckgoBarChartData`"
      }
    },
    "yAxis": {
      "type": "boolean",
      "mutable": false,
      "complexType": {
        "original": "boolean",
        "resolved": "boolean",
        "references": {}
      },
      "required": false,
      "optional": false,
      "docs": {
        "tags": [],
        "text": "If `false`, no axis y will be draw."
      },
      "attribute": "y-axis",
      "reflect": false,
      "defaultValue": "true"
    },
    "yAxisMin": {
      "type": "number",
      "mutable": false,
      "complexType": {
        "original": "number",
        "resolved": "number",
        "references": {}
      },
      "required": false,
      "optional": false,
      "docs": {
        "tags": [],
        "text": "Set a minimal value for the y Axis. Useful in case the series of data could contains only zeros."
      },
      "attribute": "y-axis-min",
      "reflect": false,
      "defaultValue": "0"
    }
  }; }
  static get events() { return [{
      "method": "chartCustomLoad",
      "name": "chartCustomLoad",
      "bubbles": true,
      "cancelable": true,
      "composed": true,
      "docs": {
        "tags": [{
            "text": undefined,
            "name": "private"
          }],
        "text": "The event to be processed to load the data if you are using a custom loader."
      },
      "complexType": {
        "original": "string",
        "resolved": "string",
        "references": {}
      }
    }, {
      "method": "chartRandomColor",
      "name": "chartRandomColor",
      "bubbles": true,
      "cancelable": true,
      "composed": true,
      "docs": {
        "tags": [],
        "text": "Emit the random colors that are generated for the charts."
      },
      "complexType": {
        "original": "string[]",
        "resolved": "string[]",
        "references": {}
      }
    }]; }
  static get methods() { return {
    "updateCurrentBar": {
      "complexType": {
        "signature": "(values: DeckdeckgoBarChartDataValue[]) => Promise<void>",
        "parameters": [{
            "tags": [{
                "text": "values",
                "name": "param"
              }],
            "text": ""
          }],
        "references": {
          "Promise": {
            "location": "global"
          },
          "DeckdeckgoBarChartDataValue": {
            "location": "import",
            "path": "@deckdeckgo/types"
          }
        },
        "return": "Promise<void>"
      },
      "docs": {
        "text": "This is the method we are using to refresh the current bar chart when an audience is participating to live vote. It will not redraw the axis but it will draw and animate the bars.",
        "tags": [{
            "name": "param",
            "text": "values"
          }]
      }
    },
    "draw": {
      "complexType": {
        "signature": "(width?: number, height?: number) => Promise<void>",
        "parameters": [{
            "tags": [{
                "text": "width",
                "name": "param"
              }],
            "text": ""
          }, {
            "tags": [{
                "text": "height",
                "name": "param"
              }],
            "text": ""
          }],
        "references": {
          "Promise": {
            "location": "global"
          }
        },
        "return": "Promise<void>"
      },
      "docs": {
        "text": "In case you would like to redraw your chart, for example on resize of the window.",
        "tags": [{
            "name": "param",
            "text": "width"
          }, {
            "name": "param",
            "text": "height"
          }]
      }
    },
    "next": {
      "complexType": {
        "signature": "() => Promise<void>",
        "parameters": [],
        "references": {
          "Promise": {
            "location": "global"
          }
        },
        "return": "Promise<void>"
      },
      "docs": {
        "text": "If you are using animation, this method is used to display the next data respectively the next chart.",
        "tags": []
      }
    },
    "prev": {
      "complexType": {
        "signature": "() => Promise<void>",
        "parameters": [],
        "references": {
          "Promise": {
            "location": "global"
          }
        },
        "return": "Promise<void>"
      },
      "docs": {
        "text": "If you are using animation, this method is used to display the previous data respectively the previous chart.",
        "tags": []
      }
    },
    "isBeginning": {
      "complexType": {
        "signature": "() => Promise<boolean>",
        "parameters": [],
        "references": {
          "Promise": {
            "location": "global"
          }
        },
        "return": "Promise<boolean>"
      },
      "docs": {
        "text": "Is animation at the begin of the serie.",
        "tags": []
      }
    },
    "isEnd": {
      "complexType": {
        "signature": "() => Promise<boolean>",
        "parameters": [],
        "references": {
          "Promise": {
            "location": "global"
          }
        },
        "return": "Promise<boolean>"
      },
      "docs": {
        "text": "Is animation at the end of the serie.",
        "tags": []
      }
    },
    "postCustomLoad": {
      "complexType": {
        "signature": "(content: string | undefined) => Promise<void>",
        "parameters": [{
            "tags": [{
                "text": "content",
                "name": "param"
              }],
            "text": ""
          }],
        "references": {
          "Promise": {
            "location": "global"
          }
        },
        "return": "Promise<void>"
      },
      "docs": {
        "text": "If you \"manually\" load the data, call this method once the text content fetched.",
        "tags": [{
            "name": "param",
            "text": "content"
          }]
      }
    }
  }; }
  static get elementRef() { return "el"; }
  static get watchers() { return [{
      "propName": "src",
      "methodName": "onSrcChange"
    }, {
      "propName": "data",
      "methodName": "onDataChange"
    }]; }
}
