import { Component, Element, Method, Prop, Watch, h, Host, Event } from '@stencil/core';
import parse from 'date-fns/parse';
import isValid from 'date-fns/isValid';
import { scaleLinear, scaleTime } from 'd3-scale';
import { extent } from 'd3-array';
import { axisBottom, axisLeft } from 'd3-axis';
import { area, curveMonotoneX } from 'd3-shape';
import 'd3-transition';
import { easeLinear } from 'd3-ease';
import { DeckdeckgoChartUtils } from '../deckdeckgo-chart';
var DeckdeckgoLineChartAxisDomain;
(function (DeckdeckgoLineChartAxisDomain) {
  DeckdeckgoLineChartAxisDomain["EXTENT"] = "extent";
  DeckdeckgoLineChartAxisDomain["MAX"] = "max";
})(DeckdeckgoLineChartAxisDomain || (DeckdeckgoLineChartAxisDomain = {}));
export class DeckdeckgoLineChart {
  constructor() {
    /**
     * The line separator use in your csv file
     */
    this.separator = ';';
    /**
     * Set to `true` in case you would like to load (fetch) the data by yourself. Useful in case your data are protected with a token.
     */
    this.customLoader = false;
    /**
     * The pattern for the dates. All supported date format: https://date-fns.org/v2.0.0-alpha.26/docs/parse.
     */
    this.datePattern = 'yyyy-MM-dd';
    /**
     * The margin top of the chart in pixel
     */
    this.marginTop = 32;
    /**
     * The margin bottom of the chart in pixel
     */
    this.marginBottom = 64;
    /**
     * The margin left of the chart in pixel
     */
    this.marginLeft = 32;
    /**
     * The margin right of the chart in pixel
     */
    this.marginRight = 32;
    /**
     * The y axis behavior.
     */
    this.yAxisDomain = DeckdeckgoLineChartAxisDomain.MAX;
    /**
     * Render smooth lines or with edges.
     */
    this.smooth = true;
    /**
     * Draw the area.
     */
    this.area = true;
    /**
     * Render a grid behind the chart.
     */
    this.grid = false;
    /**
     * Display multiple graphs and animate the transition between these
     */
    this.animation = false;
    /**
     * Duration of the transition between graphs
     */
    this.animationDuration = 1000;
    this.serieIndex = 0;
    this.randomFillColors = [];
  }
  async componentDidLoad() {
    await this.draw();
  }
  async redraw() {
    await this.draw();
  }
  /**
   * In case you would like to redraw your chart, for example on resize of the window.
   * @param width
   * @param height
   */
  draw(width, height) {
    return new Promise(async (resolve) => {
      if (width > 0) {
        this.width = width;
      }
      if (height > 0) {
        this.height = height;
      }
      if (!this.width || !this.height || !this.src) {
        resolve();
        return;
      }
      this.serieIndex = 0;
      this.svg = DeckdeckgoChartUtils.initSvg(this.el, this.width + this.marginLeft + this.marginRight, this.height + this.marginTop + this.marginBottom);
      this.svg = this.svg.append('g').attr('transform', 'translate(' + (this.marginLeft + this.marginRight) + ',' + (this.marginTop + this.marginBottom) + ')');
      this.series = await this.fetchData();
      if (!this.series || this.series.length <= 0) {
        resolve();
        return;
      }
      await this.firstDraw();
      resolve();
    });
  }
  async firstDraw() {
    await this.initAxis();
    await this.drawAxis();
    await this.drawLine(0);
    // All other series to compare if not animated
    if (this.series.length > 1 && !this.animation) {
      const promises = [];
      for (let i = 1; i < this.series.length; i++) {
        promises.push(this.drawLine(i));
      }
      await Promise.all(promises);
    }
  }
  /**
   * If you are using animation, this method is used to display the next data respectively the next chart.
   */
  async next() {
    await this.prevNext(true);
  }
  /**
   * If you are using animation, this method is used to display the previous data respectively the previous chart.
   */
  async prev() {
    await this.prevNext(false);
  }
  async prevNext(next) {
    if (!this.animation) {
      return;
    }
    if (!this.series || this.series.length <= 0) {
      return;
    }
    if (next && this.serieIndex + 1 < this.series.length) {
      this.serieIndex++;
      await this.drawLine(this.serieIndex);
    }
    else if (!next && this.serieIndex > 0) {
      this.serieIndex--;
      await this.drawLine(this.serieIndex);
    }
  }
  /**
   * Is animation at the begin of the serie.
   */
  isBeginning() {
    return new Promise((resolve) => {
      if (!this.animation) {
        resolve(true);
        return;
      }
      resolve(this.serieIndex === 0);
    });
  }
  /**
   * Is animation at the end of the serie.
   */
  isEnd() {
    return new Promise((resolve) => {
      if (!this.animation) {
        resolve(true);
        return;
      }
      resolve(this.serieIndex === this.series.length - 1);
    });
  }
  initAxis() {
    return new Promise(async (resolve) => {
      const firstSerieData = this.series[0].data;
      const isXAxisNumber = firstSerieData && firstSerieData.length > 0 && typeof firstSerieData[0].when === 'number';
      this.x = isXAxisNumber
        ? scaleLinear().range([0, this.width - this.marginLeft - this.marginRight])
        : scaleTime().range([0, this.width - this.marginLeft - this.marginRight]);
      this.y = scaleLinear().range([this.height - this.marginTop - this.marginBottom, 0]);
      this.x.domain(extent(firstSerieData, (d) => d.when));
      if (this.yAxisDomain === DeckdeckgoLineChartAxisDomain.EXTENT) {
        if (this.series.length > 1) {
          const minMaxValues = await this.findMinMaxValues();
          this.y.domain(minMaxValues);
        }
        else {
          this.y.domain(extent(firstSerieData, (d) => d.value));
        }
      }
      else {
        const minMaxValues = await this.findMinMaxValues();
        this.y.domain([0, minMaxValues[1]]);
      }
      resolve();
    });
  }
  findMinMaxValues() {
    return new Promise(async (resolve) => {
      if (!this.series || this.series.length <= 0) {
        resolve([0, 0]);
        return;
      }
      const promises = [];
      for (let i = 0; i < this.series.length; i++) {
        promises.push(this.findMinMax(i));
      }
      const results = await Promise.all(promises);
      const minMax = await this.concatMinMax(results);
      resolve([minMax.min, minMax.max]);
    });
  }
  concatMinMax(allMinMax) {
    return new Promise((resolve) => {
      if (!allMinMax || allMinMax.length <= 0) {
        resolve({
          min: 0,
          max: 0,
        });
        return;
      }
      // https://stackoverflow.com/a/30834687/5404186
      let min = allMinMax[0].min;
      let max = allMinMax[0].max;
      for (let i = 1; i < allMinMax.length; ++i) {
        if (allMinMax[i].max > max) {
          max = allMinMax[i].max;
        }
        if (allMinMax[i].min < min) {
          min = allMinMax[i].min;
        }
      }
      resolve({
        min: min,
        max: max,
      });
    });
  }
  findMinMax(index) {
    return new Promise((resolve) => {
      // https://stackoverflow.com/a/30834687/5404186
      let min = this.series[index].data[0].value;
      let max = this.series[index].data[0].value;
      for (let i = 1; i < this.series[index].data.length; ++i) {
        if (this.series[index].data[i].value > max) {
          max = this.series[index].data[i].value;
        }
        if (this.series[index].data[i].value < min) {
          min = this.series[index].data[i].value;
        }
      }
      resolve({
        min: min,
        max: max,
      });
    });
  }
  drawAxis() {
    return new Promise((resolve) => {
      const bottomAxis = this.ticks > 0 ? axisBottom(this.x).ticks(this.ticks) : axisBottom(this.x);
      const leftAxis = this.ticks > 0 ? axisLeft(this.y).ticks(this.ticks) : axisLeft(this.y);
      const styleClassAxisX = 'axis axis-x' + (this.grid ? ' axis-grid' : '');
      const styleClassAxisY = 'axis axis-y' + (this.grid ? ' axis-grid' : '');
      if (this.grid) {
        bottomAxis.tickSize(-(this.height - this.marginTop - this.marginBottom)).tickFormat(null);
        leftAxis.tickSize(-(this.width - this.marginLeft - this.marginRight)).tickFormat(null);
      }
      this.svg
        .append('g')
        .attr('class', styleClassAxisX)
        .attr('transform', 'translate(0,' + (this.height - this.marginTop - this.marginBottom) + ')')
        .call(bottomAxis)
        .selectAll('text')
        .attr('transform', 'translate(-10,0)rotate(-45)')
        .style('text-anchor', 'end');
      this.svg.append('g').attr('class', styleClassAxisY).call(leftAxis);
      resolve();
    });
  }
  drawLine(index) {
    return new Promise((resolve) => {
      let line = area().x((d) => this.x(d.when));
      if (this.area) {
        line.y0(this.height - this.marginTop - this.marginBottom).y1((d) => this.y(d.value));
      }
      else {
        line.y((d) => this.y(d.value));
      }
      if (this.smooth) {
        line.curve(curveMonotoneX);
      }
      const data = this.series[index].data;
      const randomFillColor = this.series[index].randomFillColor;
      if (this.animation) {
        this.drawAnimatedLine(data, index, line, randomFillColor);
      }
      else {
        this.drawInstantLine(data, index, line, randomFillColor);
      }
      resolve();
    });
  }
  drawInstantLine(data, index, line, randomFillColor) {
    // Random hex color source: https://css-tricks.com/snippets/javascript/random-hex-color/
    // Starting with line 1 instead of 0 is more readable. Also Pie chart style start with 1 too
    const styleIndex = index + 1;
    const background = `var(--deckgo-chart-fill-color-${styleIndex}, #${randomFillColor})`;
    this.svg
      .append('path')
      .datum(data)
      .attr('class', 'area')
      .style('fill', background)
      .style('fill-opacity', `var(--deckgo-chart-fill-opacity-${styleIndex}, 0.8)`)
      .style('stroke', `var(--deckgo-chart-stroke-${styleIndex}, var(--deckgo-chart-stroke, ${background}))`)
      .style('stroke-width', `var(--deckgo-chart-stroke-width-${styleIndex}, 3px)`)
      .attr('d', line);
  }
  drawAnimatedLine(data, index, line, randomFillColor) {
    const section = this.svg.selectAll('.area').data([data], (d) => {
      return this.y(d.value);
    });
    // Starting with line 1 instead of 0 is more readable. Also Pie chart style start with 1 too
    const styleIndex = index + 1;
    const background = `var(--deckgo-chart-fill-color-${styleIndex}, #${randomFillColor})`;
    section
      .enter()
      .append('path')
      .merge(section)
      .style('fill', background)
      .style('fill-opacity', `var(--deckgo-chart-fill-opacity-${styleIndex}, 0.8)`)
      .style('stroke', `var(--deckgo-chart-stroke-${styleIndex}, var(--deckgo-chart-stroke, ${background}))`)
      .style('stroke-width', `var(--deckgo-chart-stroke-width-${styleIndex}, 3px)`)
      .transition()
      .duration(this.animationDuration)
      .ease(easeLinear)
      .attr('class', 'area')
      .attr('d', line);
  }
  fetchData() {
    return new Promise(async (resolve) => {
      if (!this.src) {
        resolve([]);
        return;
      }
      if (this.customLoader) {
        this.chartCustomLoad.emit(this.src);
        resolve([]);
        return;
      }
      const response = await fetch(this.src);
      const content = await response.text();
      const series = await this.loadContent(content);
      resolve(series);
    });
  }
  async postCustomLoad(content) {
    this.series = await this.loadContent(content);
    if (!this.series || this.series.length <= 0) {
      return;
    }
    await this.firstDraw();
  }
  loadContent(content) {
    return new Promise(async (resolve) => {
      if (!content || content === undefined) {
        resolve([]);
        return;
      }
      const lines = content.split('\n');
      if (!lines || lines.length <= 0) {
        resolve([]);
        return;
      }
      const series = [];
      for (const line of lines) {
        const values = line.split(this.separator);
        if (values && values.length > 1) {
          const when = this.validDate(values[0]) ? parse(values[0], this.datePattern, new Date()) : parseInt(values[0]);
          if (when) {
            for (let i = 1; i < values.length; i++) {
              const value = parseInt(values[i]);
              if (!isNaN(value)) {
                // Random hex color source: https://css-tricks.com/snippets/javascript/random-hex-color/
                const randomFillColor = Math.floor(Math.random() * 16777215).toString(16);
                let data = series && series.length >= i
                  ? series[i - 1]
                  : {
                    data: [],
                    randomFillColor: this.randomFillColors[i - 1] || randomFillColor,
                  };
                if ((series === null || series === void 0 ? void 0 : series.length) < i) {
                  this.randomFillColors.push(randomFillColor);
                }
                if (!data.data || data.data.length <= 0) {
                  data.data = [];
                }
                data.data.push({
                  when: when,
                  value: parseInt(values[i]),
                });
                if (series && series.length >= i) {
                  series[i - 1] = data;
                }
                else {
                  series.push(data);
                }
              }
            }
          }
        }
      }
      this.chartRandomColor.emit(this.randomFillColors);
      resolve(series);
    });
  }
  validDate(value) {
    return this.datePattern && isValid(parse(value, this.datePattern, new Date()));
  }
  render() {
    return (h(Host, { style: { width: `${this.width}px`, height: `${this.height}px` } },
      h("svg", null)));
  }
  static get is() { return "deckgo-line-chart"; }
  static get encapsulation() { return "shadow"; }
  static get originalStyleUrls() { return {
    "$": ["deckdeckgo-line-chart.scss"]
  }; }
  static get styleUrls() { return {
    "$": ["deckdeckgo-line-chart.css"]
  }; }
  static get properties() { return {
    "width": {
      "type": "number",
      "mutable": true,
      "complexType": {
        "original": "number",
        "resolved": "number",
        "references": {}
      },
      "required": false,
      "optional": false,
      "docs": {
        "tags": [],
        "text": "The width of the chart"
      },
      "attribute": "width",
      "reflect": false
    },
    "height": {
      "type": "number",
      "mutable": true,
      "complexType": {
        "original": "number",
        "resolved": "number",
        "references": {}
      },
      "required": false,
      "optional": false,
      "docs": {
        "tags": [],
        "text": "The height of the chart"
      },
      "attribute": "height",
      "reflect": false
    },
    "src": {
      "type": "string",
      "mutable": false,
      "complexType": {
        "original": "string",
        "resolved": "string",
        "references": {}
      },
      "required": false,
      "optional": false,
      "docs": {
        "tags": [],
        "text": "The path to the source file of the data"
      },
      "attribute": "src",
      "reflect": false
    },
    "separator": {
      "type": "string",
      "mutable": false,
      "complexType": {
        "original": "string",
        "resolved": "string",
        "references": {}
      },
      "required": false,
      "optional": false,
      "docs": {
        "tags": [],
        "text": "The line separator use in your csv file"
      },
      "attribute": "separator",
      "reflect": false,
      "defaultValue": "';'"
    },
    "customLoader": {
      "type": "boolean",
      "mutable": false,
      "complexType": {
        "original": "boolean",
        "resolved": "boolean",
        "references": {}
      },
      "required": false,
      "optional": false,
      "docs": {
        "tags": [],
        "text": "Set to `true` in case you would like to load (fetch) the data by yourself. Useful in case your data are protected with a token."
      },
      "attribute": "custom-loader",
      "reflect": false,
      "defaultValue": "false"
    },
    "datePattern": {
      "type": "string",
      "mutable": false,
      "complexType": {
        "original": "string",
        "resolved": "string",
        "references": {}
      },
      "required": false,
      "optional": false,
      "docs": {
        "tags": [],
        "text": "The pattern for the dates. All supported date format: https://date-fns.org/v2.0.0-alpha.26/docs/parse."
      },
      "attribute": "date-pattern",
      "reflect": false,
      "defaultValue": "'yyyy-MM-dd'"
    },
    "marginTop": {
      "type": "number",
      "mutable": false,
      "complexType": {
        "original": "number",
        "resolved": "number",
        "references": {}
      },
      "required": false,
      "optional": false,
      "docs": {
        "tags": [],
        "text": "The margin top of the chart in pixel"
      },
      "attribute": "margin-top",
      "reflect": false,
      "defaultValue": "32"
    },
    "marginBottom": {
      "type": "number",
      "mutable": false,
      "complexType": {
        "original": "number",
        "resolved": "number",
        "references": {}
      },
      "required": false,
      "optional": false,
      "docs": {
        "tags": [],
        "text": "The margin bottom of the chart in pixel"
      },
      "attribute": "margin-bottom",
      "reflect": false,
      "defaultValue": "64"
    },
    "marginLeft": {
      "type": "number",
      "mutable": false,
      "complexType": {
        "original": "number",
        "resolved": "number",
        "references": {}
      },
      "required": false,
      "optional": false,
      "docs": {
        "tags": [],
        "text": "The margin left of the chart in pixel"
      },
      "attribute": "margin-left",
      "reflect": false,
      "defaultValue": "32"
    },
    "marginRight": {
      "type": "number",
      "mutable": false,
      "complexType": {
        "original": "number",
        "resolved": "number",
        "references": {}
      },
      "required": false,
      "optional": false,
      "docs": {
        "tags": [],
        "text": "The margin right of the chart in pixel"
      },
      "attribute": "margin-right",
      "reflect": false,
      "defaultValue": "32"
    },
    "yAxisDomain": {
      "type": "string",
      "mutable": false,
      "complexType": {
        "original": "string",
        "resolved": "string",
        "references": {}
      },
      "required": false,
      "optional": false,
      "docs": {
        "tags": [],
        "text": "The y axis behavior."
      },
      "attribute": "y-axis-domain",
      "reflect": false,
      "defaultValue": "DeckdeckgoLineChartAxisDomain.MAX"
    },
    "smooth": {
      "type": "boolean",
      "mutable": false,
      "complexType": {
        "original": "boolean",
        "resolved": "boolean",
        "references": {}
      },
      "required": false,
      "optional": false,
      "docs": {
        "tags": [],
        "text": "Render smooth lines or with edges."
      },
      "attribute": "smooth",
      "reflect": false,
      "defaultValue": "true"
    },
    "area": {
      "type": "boolean",
      "mutable": false,
      "complexType": {
        "original": "boolean",
        "resolved": "boolean",
        "references": {}
      },
      "required": false,
      "optional": false,
      "docs": {
        "tags": [],
        "text": "Draw the area."
      },
      "attribute": "area",
      "reflect": false,
      "defaultValue": "true"
    },
    "ticks": {
      "type": "number",
      "mutable": false,
      "complexType": {
        "original": "number",
        "resolved": "number",
        "references": {}
      },
      "required": false,
      "optional": false,
      "docs": {
        "tags": [],
        "text": "Render ticks on the axes."
      },
      "attribute": "ticks",
      "reflect": false
    },
    "grid": {
      "type": "boolean",
      "mutable": false,
      "complexType": {
        "original": "boolean",
        "resolved": "boolean",
        "references": {}
      },
      "required": false,
      "optional": false,
      "docs": {
        "tags": [],
        "text": "Render a grid behind the chart."
      },
      "attribute": "grid",
      "reflect": false,
      "defaultValue": "false"
    },
    "animation": {
      "type": "boolean",
      "mutable": false,
      "complexType": {
        "original": "boolean",
        "resolved": "boolean",
        "references": {}
      },
      "required": false,
      "optional": false,
      "docs": {
        "tags": [],
        "text": "Display multiple graphs and animate the transition between these"
      },
      "attribute": "animation",
      "reflect": false,
      "defaultValue": "false"
    },
    "animationDuration": {
      "type": "number",
      "mutable": false,
      "complexType": {
        "original": "number",
        "resolved": "number",
        "references": {}
      },
      "required": false,
      "optional": false,
      "docs": {
        "tags": [],
        "text": "Duration of the transition between graphs"
      },
      "attribute": "animation-duration",
      "reflect": false,
      "defaultValue": "1000"
    }
  }; }
  static get events() { return [{
      "method": "chartCustomLoad",
      "name": "chartCustomLoad",
      "bubbles": true,
      "cancelable": true,
      "composed": true,
      "docs": {
        "tags": [{
            "text": undefined,
            "name": "private"
          }],
        "text": "The event to be processed to load the data if you are using a custom loader."
      },
      "complexType": {
        "original": "string",
        "resolved": "string",
        "references": {}
      }
    }, {
      "method": "chartRandomColor",
      "name": "chartRandomColor",
      "bubbles": true,
      "cancelable": true,
      "composed": true,
      "docs": {
        "tags": [],
        "text": "Emit the random colors that are generated for the charts."
      },
      "complexType": {
        "original": "string[]",
        "resolved": "string[]",
        "references": {}
      }
    }]; }
  static get methods() { return {
    "draw": {
      "complexType": {
        "signature": "(width?: number, height?: number) => Promise<void>",
        "parameters": [{
            "tags": [{
                "text": "width",
                "name": "param"
              }],
            "text": ""
          }, {
            "tags": [{
                "text": "height",
                "name": "param"
              }],
            "text": ""
          }],
        "references": {
          "Promise": {
            "location": "global"
          }
        },
        "return": "Promise<void>"
      },
      "docs": {
        "text": "In case you would like to redraw your chart, for example on resize of the window.",
        "tags": [{
            "name": "param",
            "text": "width"
          }, {
            "name": "param",
            "text": "height"
          }]
      }
    },
    "next": {
      "complexType": {
        "signature": "() => Promise<void>",
        "parameters": [],
        "references": {
          "Promise": {
            "location": "global"
          }
        },
        "return": "Promise<void>"
      },
      "docs": {
        "text": "If you are using animation, this method is used to display the next data respectively the next chart.",
        "tags": []
      }
    },
    "prev": {
      "complexType": {
        "signature": "() => Promise<void>",
        "parameters": [],
        "references": {
          "Promise": {
            "location": "global"
          }
        },
        "return": "Promise<void>"
      },
      "docs": {
        "text": "If you are using animation, this method is used to display the previous data respectively the previous chart.",
        "tags": []
      }
    },
    "isBeginning": {
      "complexType": {
        "signature": "() => Promise<boolean>",
        "parameters": [],
        "references": {
          "Promise": {
            "location": "global"
          }
        },
        "return": "Promise<boolean>"
      },
      "docs": {
        "text": "Is animation at the begin of the serie.",
        "tags": []
      }
    },
    "isEnd": {
      "complexType": {
        "signature": "() => Promise<boolean>",
        "parameters": [],
        "references": {
          "Promise": {
            "location": "global"
          }
        },
        "return": "Promise<boolean>"
      },
      "docs": {
        "text": "Is animation at the end of the serie.",
        "tags": []
      }
    },
    "postCustomLoad": {
      "complexType": {
        "signature": "(content: string | undefined) => Promise<void>",
        "parameters": [{
            "tags": [],
            "text": ""
          }],
        "references": {
          "Promise": {
            "location": "global"
          }
        },
        "return": "Promise<void>"
      },
      "docs": {
        "text": "",
        "tags": []
      }
    }
  }; }
  static get elementRef() { return "el"; }
  static get watchers() { return [{
      "propName": "src",
      "methodName": "redraw"
    }]; }
}
