import { Component, Prop, Element, Method, Watch, h, Host, Event } from '@stencil/core';
import { pie, arc } from 'd3-shape';
import { DeckdeckgoChartUtils } from '../deckdeckgo-chart';
import 'd3-transition';
import { ascending } from 'd3-array';
export class DeckdeckgoPieChart {
  constructor() {
    /**
     * The margin top of the chart in pixel
     */
    this.marginTop = 8;
    /**
     * The margin bottom of the chart in pixel
     */
    this.marginBottom = 64;
    /**
     * The margin left of the chart in pixel
     */
    this.marginLeft = 32;
    /**
     * The margin right of the chart in pixel
     */
    this.marginRight = 32;
    /**
     * The inner radius of the pie
     */
    this.innerRadius = 0;
    /**
     * The line separator use in your csv file
     */
    this.separator = ';';
    /**
     * Set to `true` in case you would like to load (fetch) the data by yourself. Useful in case your data are protected with a token.
     */
    this.customLoader = false;
    /**
     * Display multiple graphs and animate the transition between these
     */
    this.animation = false;
    /**
     * Duration of the transition between graphs
     */
    this.animationDuration = 1000;
    this.pieDataIndex = 0;
  }
  async componentDidLoad() {
    await this.draw();
  }
  async redraw() {
    await this.draw();
  }
  /**
   * In case you would like to redraw your chart, for example on resize of the window.
   * @param width
   * @param height
   */
  draw(width, height) {
    return new Promise(async (resolve) => {
      if (width > 0) {
        this.width = width;
      }
      if (height > 0) {
        this.height = height;
      }
      if (!this.width || !this.height || !this.src) {
        resolve();
        return;
      }
      this.pieDataIndex = 0;
      const maxWidth = this.width > this.marginLeft + this.marginRight ? this.width - this.marginLeft - this.marginRight : this.width;
      const maxHeight = this.height > this.marginTop + this.marginBottom ? this.height - this.marginTop - this.marginBottom : this.height;
      this.svg = DeckdeckgoChartUtils.initSvg(this.el, maxWidth, maxHeight);
      this.svg = this.svg.append('g').attr('transform', 'translate(' + maxWidth / 2 + ',' + maxHeight / 2 + ')');
      const radius = Math.min(maxWidth, maxHeight) / 2;
      this.myPath = arc().innerRadius(this.innerRadius).outerRadius(radius);
      this.data = await this.fetchData();
      if (!this.data || this.data.length <= 0) {
        resolve();
        return;
      }
      await this.drawPie(0, 0);
      resolve();
    });
  }
  /**
   * If you are using animation, this method is used to display the next data respectively the next chart.
   */
  async next() {
    await this.prevNext(true);
  }
  /**
   * If you are using animation, this method is used to display the previous data respectively the previous chart.
   */
  async prev() {
    await this.prevNext(false);
  }
  async prevNext(next) {
    if (!this.animation) {
      return;
    }
    if (!this.data || this.data.length <= 0) {
      return;
    }
    if (next && this.pieDataIndex + 1 < this.data.length) {
      this.pieDataIndex++;
      await this.drawPie(this.pieDataIndex, this.animationDuration);
    }
    else if (!next && this.pieDataIndex > 0) {
      this.pieDataIndex--;
      await this.drawPie(this.pieDataIndex, this.animationDuration);
    }
  }
  /**
   * Is animation at the begin of the serie.
   */
  isBeginning() {
    return new Promise((resolve) => {
      if (!this.animation) {
        resolve(true);
        return;
      }
      resolve(this.pieDataIndex === 0);
    });
  }
  /**
   * Is animation at the end of the serie.
   */
  isEnd() {
    return new Promise((resolve) => {
      if (!this.animation) {
        resolve(true);
        return;
      }
      resolve(this.pieDataIndex === this.data.length - 1);
    });
  }
  async drawPie(index, animationDuration) {
    const pie = await this.createPieData(this.data[index]);
    await this.drawPieForData(pie, animationDuration);
    setTimeout(async () => {
      await this.appendLabel(pie);
    }, animationDuration);
  }
  drawPieForData(myPieData, animationDuration) {
    return new Promise((resolve) => {
      const section = this.svg.selectAll('path').data(myPieData);
      section
        .enter()
        .append('path')
        .merge(section)
        .attr('style', (d) => {
        const randomColor = d.data.randomFillColor ? `, #${d.data.randomFillColor}` : '';
        const background = `var(--deckgo-chart-fill-color-${d.data.key}${randomColor})`;
        return ('fill: ' +
          background +
          '; fill-opacity: var(--deckgo-chart-fill-opacity-' +
          d.data.key +
          ', 0.8); stroke: var(--deckgo-chart-stroke-' +
          d.data.key +
          ', ' +
          background +
          '); stroke-width: var(--deckgo-chart-stroke-width-' +
          d.data.key +
          ', 3px)');
      })
        .transition()
        .duration(animationDuration)
        .attr('d', this.myPath);
      section.exit().remove();
      resolve();
    });
  }
  async createPieData(data) {
    return new Promise(async (resolve) => {
      // Generate a pie chart
      const myPie = pie()
        .value((d) => {
        return d.value;
      })
        .sort((a, b) => {
        return ascending(a.label, b.label);
      });
      const result = myPie(data.values);
      resolve(result);
    });
  }
  appendLabel(myPieData) {
    return new Promise((resolve) => {
      this.svg.selectAll('.text-arc').remove();
      // Create a special arcs for the labels in order to always display them above the pie's slices
      const labelArcs = this.svg
        .selectAll('.text-arc')
        .data(myPieData)
        .enter()
        .append('g')
        .attr('class', 'text-arc');
      const text = labelArcs
        .append('text')
        .attr('transform', (d) => {
        return 'translate(' + this.myPath.centroid(d) + ')';
      })
        .attr('dy', '.35em')
        .style('text-anchor', 'middle');
      text
        .append('tspan')
        .attr('x', 0)
        .attr('y', '-0.7em')
        .style('font-weight', 'bold')
        .text((d) => {
        return d.data.label ? d.data.label : '';
      });
      text
        .filter((d) => d.endAngle - d.startAngle > 0.25)
        .append('tspan')
        .attr('x', 0)
        .attr('y', '0.7em')
        .text((d) => {
        return d.data.value + '%';
      });
      resolve();
    });
  }
  fetchData() {
    return new Promise(async (resolve) => {
      if (!this.src) {
        resolve([]);
        return;
      }
      if (this.customLoader) {
        this.chartCustomLoad.emit(this.src);
        resolve([]);
        return;
      }
      const response = await fetch(this.src);
      const content = await response.text();
      let results = await this.loadContent(content);
      resolve(results);
    });
  }
  async postCustomLoad(content) {
    this.data = await this.loadContent(content);
    if (!this.data || this.data.length <= 0) {
      return;
    }
    await this.drawPie(0, 0);
  }
  loadContent(content) {
    return new Promise(async (resolve) => {
      if (!content || content === undefined) {
        resolve([]);
        return;
      }
      const lines = content.split('\n');
      if (!lines || lines.length <= 0) {
        resolve([]);
        return;
      }
      let results = [];
      lines.forEach((line, lineIndex) => {
        var _a;
        const values = line.split(this.separator);
        if (values && values.length >= 2) {
          this.initRandomColors(lines);
          const label = values[0];
          const pieData = {
            label: label,
            value: parseInt(values[1]),
            randomFillColor: ((_a = this.randomColors) === null || _a === void 0 ? void 0 : _a[lineIndex]) || undefined,
            key: lineIndex + 1,
          };
          if (!isNaN(pieData.value)) {
            if (results.length <= 0) {
              results.push({
                values: [],
              });
            }
            results[0].values.push(pieData);
          }
          if (values.length > 2) {
            for (let i = 2; i < values.length; i++) {
              const tmp = parseInt(values[i]);
              if (!isNaN(tmp)) {
                if (results.length < i) {
                  results.push({
                    values: [],
                  });
                }
                const pieData = {
                  label: label,
                  value: parseInt(values[i]),
                  randomFillColor: this.randomColors.length >= i ? this.randomColors[lineIndex] : undefined,
                  key: lineIndex + 1,
                };
                results[i - 1].values.push(pieData);
              }
            }
          }
        }
      });
      resolve(results);
    });
  }
  initRandomColors(lines) {
    var _a;
    if (((_a = this.randomColors) === null || _a === void 0 ? void 0 : _a.length) === lines.length) {
      return;
    }
    this.randomColors = Array.from({ length: lines.length }, (_v, _i) => Math.floor(Math.random() * 16777215).toString(16));
    this.chartRandomColor.emit(this.randomColors);
  }
  render() {
    return (h(Host, { style: { width: `${this.width}px`, height: `${this.height}px` } },
      h("svg", null)));
  }
  static get is() { return "deckgo-pie-chart"; }
  static get encapsulation() { return "shadow"; }
  static get originalStyleUrls() { return {
    "$": ["deckdeckgo-pie-chart.scss"]
  }; }
  static get styleUrls() { return {
    "$": ["deckdeckgo-pie-chart.css"]
  }; }
  static get properties() { return {
    "width": {
      "type": "number",
      "mutable": true,
      "complexType": {
        "original": "number",
        "resolved": "number",
        "references": {}
      },
      "required": false,
      "optional": false,
      "docs": {
        "tags": [],
        "text": "The width of the chart"
      },
      "attribute": "width",
      "reflect": false
    },
    "height": {
      "type": "number",
      "mutable": true,
      "complexType": {
        "original": "number",
        "resolved": "number",
        "references": {}
      },
      "required": false,
      "optional": false,
      "docs": {
        "tags": [],
        "text": "The height of the chart"
      },
      "attribute": "height",
      "reflect": false
    },
    "marginTop": {
      "type": "number",
      "mutable": false,
      "complexType": {
        "original": "number",
        "resolved": "number",
        "references": {}
      },
      "required": false,
      "optional": false,
      "docs": {
        "tags": [],
        "text": "The margin top of the chart in pixel"
      },
      "attribute": "margin-top",
      "reflect": false,
      "defaultValue": "8"
    },
    "marginBottom": {
      "type": "number",
      "mutable": false,
      "complexType": {
        "original": "number",
        "resolved": "number",
        "references": {}
      },
      "required": false,
      "optional": false,
      "docs": {
        "tags": [],
        "text": "The margin bottom of the chart in pixel"
      },
      "attribute": "margin-bottom",
      "reflect": false,
      "defaultValue": "64"
    },
    "marginLeft": {
      "type": "number",
      "mutable": false,
      "complexType": {
        "original": "number",
        "resolved": "number",
        "references": {}
      },
      "required": false,
      "optional": false,
      "docs": {
        "tags": [],
        "text": "The margin left of the chart in pixel"
      },
      "attribute": "margin-left",
      "reflect": false,
      "defaultValue": "32"
    },
    "marginRight": {
      "type": "number",
      "mutable": false,
      "complexType": {
        "original": "number",
        "resolved": "number",
        "references": {}
      },
      "required": false,
      "optional": false,
      "docs": {
        "tags": [],
        "text": "The margin right of the chart in pixel"
      },
      "attribute": "margin-right",
      "reflect": false,
      "defaultValue": "32"
    },
    "innerRadius": {
      "type": "number",
      "mutable": false,
      "complexType": {
        "original": "number",
        "resolved": "number",
        "references": {}
      },
      "required": false,
      "optional": false,
      "docs": {
        "tags": [],
        "text": "The inner radius of the pie"
      },
      "attribute": "inner-radius",
      "reflect": false,
      "defaultValue": "0"
    },
    "src": {
      "type": "string",
      "mutable": false,
      "complexType": {
        "original": "string",
        "resolved": "string",
        "references": {}
      },
      "required": false,
      "optional": false,
      "docs": {
        "tags": [],
        "text": "The path to the source file of the data"
      },
      "attribute": "src",
      "reflect": false
    },
    "separator": {
      "type": "string",
      "mutable": false,
      "complexType": {
        "original": "string",
        "resolved": "string",
        "references": {}
      },
      "required": false,
      "optional": false,
      "docs": {
        "tags": [],
        "text": "The line separator use in your csv file"
      },
      "attribute": "separator",
      "reflect": false,
      "defaultValue": "';'"
    },
    "customLoader": {
      "type": "boolean",
      "mutable": false,
      "complexType": {
        "original": "boolean",
        "resolved": "boolean",
        "references": {}
      },
      "required": false,
      "optional": false,
      "docs": {
        "tags": [],
        "text": "Set to `true` in case you would like to load (fetch) the data by yourself. Useful in case your data are protected with a token."
      },
      "attribute": "custom-loader",
      "reflect": false,
      "defaultValue": "false"
    },
    "animation": {
      "type": "boolean",
      "mutable": false,
      "complexType": {
        "original": "boolean",
        "resolved": "boolean",
        "references": {}
      },
      "required": false,
      "optional": false,
      "docs": {
        "tags": [],
        "text": "Display multiple graphs and animate the transition between these"
      },
      "attribute": "animation",
      "reflect": false,
      "defaultValue": "false"
    },
    "animationDuration": {
      "type": "number",
      "mutable": false,
      "complexType": {
        "original": "number",
        "resolved": "number",
        "references": {}
      },
      "required": false,
      "optional": false,
      "docs": {
        "tags": [],
        "text": "Duration of the transition between graphs"
      },
      "attribute": "animation-duration",
      "reflect": false,
      "defaultValue": "1000"
    }
  }; }
  static get events() { return [{
      "method": "chartCustomLoad",
      "name": "chartCustomLoad",
      "bubbles": true,
      "cancelable": true,
      "composed": true,
      "docs": {
        "tags": [{
            "text": undefined,
            "name": "private"
          }],
        "text": "The event to be processed to load the data if you are using a custom loader."
      },
      "complexType": {
        "original": "string",
        "resolved": "string",
        "references": {}
      }
    }, {
      "method": "chartRandomColor",
      "name": "chartRandomColor",
      "bubbles": true,
      "cancelable": true,
      "composed": true,
      "docs": {
        "tags": [],
        "text": "Emit the random colors that are generated for the charts."
      },
      "complexType": {
        "original": "string[]",
        "resolved": "string[]",
        "references": {}
      }
    }]; }
  static get methods() { return {
    "draw": {
      "complexType": {
        "signature": "(width?: number, height?: number) => Promise<void>",
        "parameters": [{
            "tags": [{
                "text": "width",
                "name": "param"
              }],
            "text": ""
          }, {
            "tags": [{
                "text": "height",
                "name": "param"
              }],
            "text": ""
          }],
        "references": {
          "Promise": {
            "location": "global"
          }
        },
        "return": "Promise<void>"
      },
      "docs": {
        "text": "In case you would like to redraw your chart, for example on resize of the window.",
        "tags": [{
            "name": "param",
            "text": "width"
          }, {
            "name": "param",
            "text": "height"
          }]
      }
    },
    "next": {
      "complexType": {
        "signature": "() => Promise<void>",
        "parameters": [],
        "references": {
          "Promise": {
            "location": "global"
          }
        },
        "return": "Promise<void>"
      },
      "docs": {
        "text": "If you are using animation, this method is used to display the next data respectively the next chart.",
        "tags": []
      }
    },
    "prev": {
      "complexType": {
        "signature": "() => Promise<void>",
        "parameters": [],
        "references": {
          "Promise": {
            "location": "global"
          }
        },
        "return": "Promise<void>"
      },
      "docs": {
        "text": "If you are using animation, this method is used to display the previous data respectively the previous chart.",
        "tags": []
      }
    },
    "isBeginning": {
      "complexType": {
        "signature": "() => Promise<boolean>",
        "parameters": [],
        "references": {
          "Promise": {
            "location": "global"
          }
        },
        "return": "Promise<boolean>"
      },
      "docs": {
        "text": "Is animation at the begin of the serie.",
        "tags": []
      }
    },
    "isEnd": {
      "complexType": {
        "signature": "() => Promise<boolean>",
        "parameters": [],
        "references": {
          "Promise": {
            "location": "global"
          }
        },
        "return": "Promise<boolean>"
      },
      "docs": {
        "text": "Is animation at the end of the serie.",
        "tags": []
      }
    },
    "postCustomLoad": {
      "complexType": {
        "signature": "(content: string | undefined) => Promise<void>",
        "parameters": [{
            "tags": [],
            "text": ""
          }],
        "references": {
          "Promise": {
            "location": "global"
          }
        },
        "return": "Promise<void>"
      },
      "docs": {
        "text": "",
        "tags": []
      }
    }
  }; }
  static get elementRef() { return "el"; }
  static get watchers() { return [{
      "propName": "src",
      "methodName": "redraw"
    }]; }
}
